#include<stdio.h>
#include<stdlib.h>
#include <gmp.h>
#include <mpfr.h>
#include <math.h>

union intfloat {
	int i;
	float f;
};


int example_triangle() {
	// Area of triangle
	float a, b, c;
	a = 10.1f;
	b = 5.05f;
	
	FILE *fp;
	fp=fopen("cancellation.csv", "w");
	if(fp == NULL)
		exit(-1);
	fprintf(fp, "c, relative error of float, relative error of (s-a) in float, relative error of stable formula\n");
	// Use normal formula and mpfr 
	mpfr_t ma, mb, mc, ms, tmp1, tmp2, marea, motherarea, sminusa;
	mpfr_init2 (ma, 24);
	mpfr_init2 (mb, 24);
	mpfr_init2 (mc, 24);
	mpfr_init2 (motherarea, 100);
	mpfr_init2 (ms, 300);
	mpfr_init2 (tmp1, 300);
	mpfr_init2 (sminusa, 300);
	mpfr_init2 (tmp2, 1300);
	mpfr_init2 (marea, 1300);
	
	mpfr_set_d (ma, a, MPFR_RNDN);
	mpfr_set_d (mb, b, MPFR_RNDN);

	for (int i=1; i<1000; i++) {
		c = 5.05f + (i * 0.000000238f); // 2**-22 
		mpfr_set_d (mc, c, MPFR_RNDN);
		
		// Unstable computation
		// All this should be exact
		mpfr_add(ms, ma, mb, MPFR_RNDN);
		mpfr_add(ms, ms, mc, MPFR_RNDN);
		mpfr_div_ui(ms, ms, 2, MPFR_RNDN);
		mpfr_sub(sminusa, ms, ma, MPFR_RNDN);
		mpfr_mul(tmp2, ms, sminusa, MPFR_RNDN);
		mpfr_sub(tmp1, ms, mb, MPFR_RNDN);
		mpfr_mul(tmp2, tmp2, tmp1, MPFR_RNDN);
		mpfr_sub(tmp1, ms, mc, MPFR_RNDN);
		mpfr_mul(tmp2, tmp2, tmp1, MPFR_RNDN);
		
		// Inexact but on 1300 bits so should be fine
		mpfr_sqrt(marea, tmp2, MPFR_RNDN);

		// normal computation in float
		float s, area;
		s= 0.5f * (a + b + c);
		area = sqrtf( s * (s-a) * (s-b) * (s-c));

		mpfr_set_d(motherarea, area, MPFR_RNDN);
		mpfr_sub(tmp2, marea, motherarea, MPFR_RNDN);
		mpfr_abs(tmp2, tmp2, MPFR_RNDN);
		mpfr_div(tmp2, tmp2, marea, MPFR_RNDN);

		float float_error = mpfr_get_flt(tmp2, MPFR_RNDN);
		
		mpfr_set_d(tmp2, s-a, MPFR_RNDN);
		mpfr_sub(tmp2, sminusa, tmp2, MPFR_RNDN);
		mpfr_abs(tmp2, tmp2, MPFR_RNDN);
		mpfr_div(tmp2, tmp2, sminusa, MPFR_RNDN);

		float float_error_sminusa = mpfr_get_flt(tmp2, MPFR_RNDN);
		
		/* This doesn't show much, it's still an issue but at a different scale
		// normal computation in double
		double ds, darea, da, db, dc;
		da = (double) a;
		db = (double) b;
		dc = (double) c;
		ds= 0.5 * (da + db + dc);
		darea = sqrt( ds * (ds-a) * (ds-b) * (ds-c));
		
		mpfr_set_d(motherarea, darea, MPFR_RNDN);
		mpfr_sub(tmp2, marea, motherarea, MPFR_RNDN);
		mpfr_abs(tmp2, tmp2, MPFR_RNDN);
		mpfr_div(tmp2, tmp2, marea, MPFR_RNDN);

		float double_error = mpfr_get_flt(tmp2, MPFR_RNDN);
		*/
		
		// stable formula in float
		float stable_area;
		stable_area = 0.25f * sqrtf( (a+(b+c)) * (c-(a-b)) * (c+(a-b)) * (a + (b-c)));
		
		mpfr_set_d(motherarea, stable_area, MPFR_RNDN);
		mpfr_sub(tmp2, marea, motherarea, MPFR_RNDN);
		mpfr_abs(tmp2, tmp2, MPFR_RNDN);
		mpfr_div(tmp2, tmp2, marea, MPFR_RNDN);

		float stable_error = mpfr_get_flt(tmp2, MPFR_RNDN);
		
		fprintf(fp, "\"%2.6f\" , \"%2.6f\", \"%2.6f\" , \"%2.6f\"\n", c, float_error, float_error_sminusa, stable_error);
	}
	fclose(fp);
	return 0;
}



int example_exp() {
	// exponential
	float a, b;
	
	// Use normal formula and mpfr 
	mpfr_t ma, mb,tmp1, tmp2, mexp, motherexp;
	mpfr_init2 (ma, 24);
	mpfr_init2 (mb, 24);
	mpfr_init2 (motherexp, 100);
	mpfr_init2 (tmp1, 300);
	mpfr_init2 (tmp2, 1300);
	mpfr_init2 (mexp, 1300);
	
	float array_a[5] = {2.0f, 27.0f, -600.0f, -130.0f, -102.0f}; 	
	float array_b[5] = {9.0f, 66.0f, 600.0f, 40.0f, 75.0f}; 	
	
	for (int i=0; i<5; i++) {
		a = array_a[i];
		b = array_b[i];
		// float sum before
		float float_exp = expf( a+b );

		// float sum after
		float float2_exp = expf( a) * expf(b );
		
		printf("a=%2.0f, b=%2.0f,\t e^(a+b)=%0.6e, e^a * e^b = %0.6e\n", a, b, float_exp, float2_exp);
		getchar();
	}
	
	FILE *fp;
	fp=fopen("exponential.csv", "w");
	if(fp == NULL)
		exit(-1);
	fprintf(fp, "b, difference between the two results\n");
	
	a = 75.0f;	
	float a2 = 24.6f;	
	float a3 = 2.0f;	
	b = -105.0f;
	
	for (int i=0; i<65; i++) {
		float b_loop = b + i*0.125f;
		// float sum before
		float float_exp = expf( a+b_loop );

		// float sum after
		float float2_exp = expf( a) * expf(b_loop );
		
		double difference = fabs(((double) float_exp - float2_exp)/float_exp);
		
		// float sum before
		float float_exp2 = expf( a2+b_loop );

		// float sum after
		float float2_exp2 = expf( a2) * expf(b_loop );
		
		double difference2 = fabs(((double) float_exp2 - float2_exp2)/float_exp2);
		
		// float sum before
		float float_exp3 = expf( a3+b_loop );

		// float sum after
		float float2_exp3 = expf( a3) * expf(b_loop );
		
		double difference3 = fabs(((double) float_exp3 - float2_exp3)/float_exp3);
		
		fprintf(fp, "\"%5.3f\" , \"%2.6f\", \"%2.6f\", \"%2.6f\"\n", b_loop, difference, difference2, difference3);

	}
	fclose(fp);
	
	return 0;
}


int count_time() {
	// increments of 100ms
	float float_hour = 0.0f;
	int int_hour = 0;

	for (int i =0; i < 36000; i++) {
		float_hour += 0.1f;
		int_hour += 100;
	}

	float int_hour_sec = int_hour/ 1000.0f;
	double error = fabs(((double) float_hour-int_hour_sec));

	printf("One hour counted in float is %2.2f seconds and counted in int is %2.0f seconds. The error is %2.2f seconds\n", float_hour, int_hour_sec, error);
	
	for (int i =0; i < 36000*23; i++) {
		float_hour += 0.1f;
		int_hour += 100;
	}
	
	int_hour_sec = int_hour/ 1000.0f;
	error = fabs(((double) float_hour-int_hour_sec));

	printf("One day counted in float is %2.2f seconds and counted in int is %2.0f seconds. The error is %2.2f seconds\n", float_hour, int_hour_sec, error);
	
	for (int i =0; i < 36000*144; i++) {
		float_hour += 0.1f;
		int_hour += 100;
	}
	
	int_hour_sec = int_hour/ 1000.0f;
	error = fabs(((double) float_hour-int_hour_sec));

	printf("One week counted in float is %2.2f seconds and counted in int is %2.0f seconds. The error is %2.2f seconds\n", float_hour, int_hour_sec, error);
	
	for (int i =0; i < 36000*576; i++) {
		float_hour += 0.1f;
		int_hour += 100;
	}
	
	int_hour_sec = int_hour/ 1000.0f;
	error = fabs(((double) float_hour-int_hour_sec));

	printf("One month counted in float is %2.2f seconds and counted in int is %2.0f seconds. The error is %2.2f seconds\n", float_hour, int_hour_sec, error);
	return 0;
}


int for_loop() {
	printf("For loop from 0 to 16777217 with integer\n");
	for (int i = 0; i<16777217; i+=1) {
		if (i >= 16777215 || i % 1000000 == 0) {
			printf("%i ", i);
		}
	}

	printf("\nFor loop for 0 to 16777217 with float\n");
	
	printf("Press enter to try with float\n");
	getchar();
	
	int error_count = 0;
	for (float i = 0.0; i<16777217.0; i+=1.0) {
		if (i >= 16777215 || (int) i % 1000000 == 0) {
			printf("%2.1f ", i);
		}
		if (i == (i + 1.0f)) { 
			error_count ++;
			if (error_count > 100)  {
				break;
			}
		}
	}
	printf("^C\n");

	return 0;
}

int main() {
	printf("┌─────────────────────────────┐\n");
	printf("│         Range issues        │\n");
	printf("└─────────────────────────────┘\n");
	printf("Press enter to continue\n");
	getchar();
	example_exp();
	// show plot
	system("gnuplot -p exponential.gnuplot");
	printf("┌─────────────────────────────┐\n");
	printf("│    Representation issues    │\n");
	printf("└─────────────────────────────┘\n");
	printf("Press enter to continue\n");
	getchar();
	count_time();
	printf("┌─────────────────────────────┐\n");
	printf("│       Rounding issues       │\n");
	printf("└─────────────────────────────┘\n");
	printf("Press enter to continue\n");
	getchar();
	for_loop();
	printf("┌─────────────────────────────┐\n");
	printf("│     Cancellation issues     │\n");
	printf("└─────────────────────────────┘\n");
	printf("Press enter to continue\n");
	getchar();
	example_triangle();
	// show plot
	system("gnuplot -p cancellation.gnuplot");
	return 0;
}
